open Games
open Hashtbl

let small_float = 0.001
let k dt dd = (2.*.dt*.sqrt(dt)/.dd) +. 70.

type 'a tree = Node of 'a * ('a tree) list

type moveinfo = 
    {
      time : int;   (* when did the move appear ? *)
      j_time : int; (* when did the justifier of the move appeared ? *)
      lab : label;    (* label of the move *)
      comp : int;   (* component of the move *)
      mutable succ : int list; (* list of the moves justified by myself *)
      color : bool; (* Do we want the move to be colored ? *)
      mutable address : int list (* The address of the move in the formula *)
    }



let string_of_lab = function 
  | Question i -> "_{"^(string_of_int i)^"}"
  | Answer k -> string_of_int k

(* Here we try to get a representation of a play. *)

(* First, the prepare function organizes the work pointers as inverses to the previous pointers, and 
 * returns the first move of the play *)
let prepare p = 
  let rec aux m = match m.previous with 
  | None -> Some m
  | Some(m') -> 
      m'.work <- Some(m);
      aux m'
  in match p.current with 
    | None -> None
    | Some p_current -> aux p_current

exception Play_too_big 

let rec intlabel_of_move m = match m.label with
  | Question i -> i
  | Answer _ -> 
      (match m.justifier with
	 | None -> failwith "Unjustified answer !"
	 | Some m' -> intlabel_of_move m')

(* La fonction suivante prend en argument une forêt donnant la largeur des adresses présentes, et la met
à jour avec l'adjonction d'une nouvelle adresse *)

let add_preaddr f addr =
  let rec aux l addr = match (l, addr) with
    | _, [] -> assert false
    | Node((n,true),f)::q, [0] -> (0, Node((n,true),f)::q)
    | Node((n,false),f)::q, [0] -> (1, Node((n+1, true), f)::q)
    | (Node((p,b), f))::q, 0::ad -> 
	let (n, f') = aux f ad in (n, Node((p+n,b), f')::q)
    | t::q, i::ad -> 
	let (n, q') = aux q ((i-1)::ad) in (n, t::q')
    | [], [0] -> (1, [Node((1,true), [])])
    | [], 0::ad -> 
	let (n, f') = aux [] ad in
	  (n, [Node((n,false), f')])
    | [], i::ad -> 
	let (n, q') = aux [] ((i-1)::ad) in
	  (n, Node((0,false),[])::q') in
  let (n, f') = aux f addr in f'	  

(* La fonction suivante prend la forêt des adresses en argument, et renvoie sa largeur totale *)

let width = List.fold_left (fun n t -> match t with Node(p, _) -> n+p) 0

(* La fonction suivante prend en argument un arbre d'adresses, une adresse, et donne en réponse l'abscisse 
de l'argument dans la représentation linéaire de la formule *)

let rec addr_of_preaddr = function
    | t::q -> ((treeaddr_of_preaddr t)::(addr_of_preaddr q))
    | [] -> []
and treeaddr_of_preaddr = function
    | Node((n,b),f) -> Node(n, addr_of_preaddr f)


let addr_to_depth f addr = 
  let rec aux f addr =
    match (f, addr) with
      | [], _ -> failwith "Adresse donnée non présente dans l'arbre des adresses !"
      | _, [] -> failwith "Ce cas n'existe pas!"
      | Node(n,f')::l, [0] -> n-1
      | Node(n,f')::_, 0::l -> aux f' l
      | Node(n,_)::f', i::l -> n + (aux f' ((i-1)::l))
  in ((width f) - (aux f addr) - 1)





let rec string_of_tree = function
	| Node(_, []) -> "o"
	| Node(_, [Node(_, [])]) -> "o->o"
	| Node(_, l) -> ((string_of_list l)^"->o")
and string_of_list = function
	| [] -> ""
	| [Node(_,[])] -> "o"
	| [t] -> ("("^(string_of_tree t)^")")
	| Node(0, _)::q -> string_of_list q
(*	| Node(_,[])::q -> ("o -> "^(string_of_list q))*)
(*	| [t1; Node(_,[])] -> ("("^(string_of_tree t1)^") -> o")*)
	| t::q -> ("("^(string_of_tree t)^")->"^(string_of_list q))

let formula_of_addresses = function
  | [Node(0, _); Node(0, _); Node(_, [Node(_,c2)])] -> ((string_of_list c2)^" -> o")
  | [Node(0, _); Node(_, [Node(_,c1)]); Node(_, [Node(_, c2)])] -> ((string_of_list c1)^" -> o ==> "^(string_of_list c2)^" -> o")
  | [Node(_, l0); Node(_, l1); Node(_, [Node(_, c2)])] -> ((string_of_list l0)^" ==> "^(string_of_list l1)^" ==> "^(string_of_list c2)^" -> o")
  | l -> failwith ("Invalid components. List has length "^(string_of_int (List.length (match (List.nth l 1) with Node(_,l) -> l))))


let string_of_address addr = 
  let rec aux = function
    | [] -> "]"
    | [n] -> ((string_of_int n)^"]")
    | n::l -> ((string_of_int n)^";"^(aux l))
in ("["^(aux addr))
	


(*The next function takes a prepared play, and creates a hashtable which associates to each move its moveinfo*)

let play_to_hashtbl p = 
  let n = Games.length p in 
   (* (if  n > 577 then raise Play_too_big else ());*)
    let h = Hashtbl.create n in 
    let rec aux m k = 
      let t = k
      and j = match m.justifier with
	| None -> k
	| Some(m') -> let minfo = Hashtbl.find h m' in minfo.time
      and l = m.label
      and c = m.component
      and addr = 
	(
	  match m.label with
	    | Question _ -> 
		(match m.justifier with
		   | None -> [(intlabel_of_move m); (2-m.component)]
		   | Some m' -> 
		       if m.component = m'.component then
			 let minfo = Hashtbl.find h m' in
			   (intlabel_of_move m)::(minfo.address)
		       else [(intlabel_of_move m); (2-m.component)]
		)
	    | Answer _ -> 
		(
		  match m.justifier with
		    | None -> failwith "Unjustified answer !"
		    | Some m' -> let minfo = Hashtbl.find h m' in minfo.address
		)
	)
      in
      	(
	  match m.justifier with 
	    | None -> ()
	    | Some m' -> let minfo = Hashtbl.find h m' in 
		minfo.succ <- (t::minfo.succ)
	);
	Hashtbl.add h m
	  {
	    time = t;
	    j_time = j;
	    lab = l;
	    comp = c;
	    succ = [];
	    color = false;
	    address = addr
	  };
	(match m.work with 
	   | None -> ()
	   | Some(m') -> aux m' (k+1))
    and i = 
      (match prepare p with 
	 | None -> failwith "Cannot convert empty play to string !"
	 | Some x -> x )
    in
      aux i 0;
      h 

let play_to_hashtbl_colors p = 
  let n = Games.length p in 
    (if  n > 577 then raise Play_too_big else ());
    let h = Hashtbl.create n 
    and i = 
      match prepare p with 
	| None -> failwith "Cannot convert empty play to string !"
	| Some x -> x in
    let h_views = Misc.compute_views p in
    let h_colors = Hashtbl.create n in
    let rec aux m k = 
      let t = k
      and j = match m.justifier with
	| None -> k
	| Some(m') -> let minfo = Hashtbl.find h m' in minfo.time
      and l = m.label
      and c = m.component 
      and colors = 
    	let (v,_) = Hashtbl.find h_views m in
	  if Hashtbl.mem h_colors v then true else
	    (Hashtbl.add h_colors v ();false)
      and addr = match m.justifier with
	| None -> [intlabel_of_move m; (2-m.component)]
	| Some m' -> 
	    if m.component = m'.component then
	      let minfo = Hashtbl.find h m' in
		(intlabel_of_move m)::(minfo.address)
	    else [(intlabel_of_move m); (2-m.component)] in
	(
	  match m.justifier with 
	    | None -> ()
	    | Some m' -> let minfo = Hashtbl.find h m' in 
		minfo.succ <- (t::minfo.succ)
	);
	Hashtbl.add h m
	  {
	    time = t;
	    j_time = j;
	    lab = l;
	    comp = c;
	    succ = [];
	    color = colors;
	    address = addr
	  };
	
	match m.work with 
	  | None -> ()
	  | Some(m') -> aux m' (k+1) in
      aux i 0;
      h




let up_point dt dd= 
  (string_of_float (max (dd/.2. -. dt/.(k dt dd)) 0.))^"r"^(string_of_float (max (dt/.2.-.dd/.(k dt dd)) 0.))^"u"

let rec down_points n0 time_to_depth depth = function
  | [] -> ""
  | [n] -> 
      let d0 = depth
      and d = (time_to_depth.(n)) in 
	let dt = float_of_int (n - n0)
	and dd = float_of_int (d - d0) in
	(string_of_float (dd/.2. +. dt/.(k dt dd)))^"l"^(string_of_float (dt/.2.+.dd/.(k dt dd)))^"d"
  | n::q -> 
      let d0 = depth
      and d = (time_to_depth.(n)) in
	let dt = float_of_int (n - n0)
	and dd = float_of_int (d - d0) in
	(string_of_float (dd/.2. +. dt/.(k dt dd)))^"l"^(string_of_float (dt/.2.+.dd/.(k dt dd)))^"d"^","
	^(down_points n0 time_to_depth depth q)

let formula h = 
  Hashtbl.iter (fun _ info -> (info.address <- List.rev (info.address))) h;
  let preaddresses = Hashtbl.fold
    (fun _ cur f -> add_preaddr f (cur.address)) h [] in
  let addresses = addr_of_preaddr preaddresses in
    formula_of_addresses addresses
      
let hashtbl_to_vector h =
  Hashtbl.iter (fun _ info -> (info.address <- List.rev (info.address))) h;
  let length = Hashtbl.fold 
    (fun _ cur t -> (max t cur.time)) h 0
  and preaddresses = Hashtbl.fold 
    (fun _ cur f -> add_preaddr f (cur.address)) h [] in
  let addresses = addr_of_preaddr preaddresses in
    print_endline ("Playing on formula "^(formula_of_addresses addresses));
    (*	print_endline ("Length of the play is : "^((string_of_int (length+1))));*)
    let max_depth = width addresses in
    let v = Array.make_matrix (length+1) (max_depth+1) ""
    and time_to_depth = Array.make (length+1) 0 in
      (*	print_endline ("Length of time_to_depth is : "^(string_of_int (Array.length time_to_depth)));*)
      Hashtbl.iter (fun _ info -> 
		      let depth = addr_to_depth addresses (info.address) in 
			time_to_depth.(info.time) <- depth;
	(*		print_endline ("Le coup "^(string_of_int (info.time))^" a pour adresse "^(string_of_address (info.address)))*)) h;
      Hashtbl.iter
	(fun _ info -> 
	   ( let depth = time_to_depth.(info.time) in
	       (*	print_endline ("time_to_depth.("^(string_of_int (info.time))^") <- "^(string_of_int depth)^";");*)
	       v.(info.time).(max_depth - depth) <- 
		 (
		   (
		     let s = 
		       (
			 match (info.time mod 2) with 
			   | 0 -> 
		       	       (
				 match info.lab with 
                                   | Question _ -> if info.color then 
				       "\\color{red}{\\circ"^(string_of_lab info.lab)^"}"
				     else
				       "\\circ"
					 (*^(string_of_lab info.lab)^":"^(string_of_int info.comp)*)
				   | Answer _ -> if info.color then 
				       "\\color{red}{"^(string_of_lab info.lab)^"}" (*^":"^(string_of_int info.comp)*)
				     else
				       (string_of_lab info.lab)(*^":"^(string_of_int info.comp)*)
			       )
	       		   | _ -> 
			       (
				 match info.lab with 
                                   | Question _ -> if info.color then 
				       "\\color{red}{\\bullet"^(string_of_lab info.lab)^"}"
				     else
				       "\\bullet"
					 (*^(string_of_lab info.lab)^":"^(string_of_int info.comp)*)
				   | Answer _ -> if info.color then 
				       "\\color{red}{"^(string_of_lab info.lab)^"}" (*^":"^(string_of_int info.comp)*)
				     else
				       (string_of_lab info.lab)(*^":"^(string_of_int info.comp)*)
			       )
		       ) in 
		       match (info.time=info.j_time) with 
			 | true -> "0box"^(string_of_int (List.length info.succ))^"("^
			     (
			       match List.length info.succ with 
				 | 0 -> ")[l,t=#$"^s^"$#,b=0]"
				 | _ -> (down_points info.time time_to_depth (time_to_depth.(info.time)) (info.succ))^")"^"[l,t=#$"^s^"$#,b=0]"
			     )
			 | false -> 
			     let t = info.time
			     and t0 = info.j_time in
			     let d = time_to_depth.(t)
			     and d0 = time_to_depth.(t0) in
			     let dd = float_of_int (d-d0)
			     and dt = float_of_int (t-t0) in
			       (*	print_endline ("Point n°"^(string_of_int t)^" is pointing to time "^(string_of_int t0)^" at depth "^(string_of_int d0)^" right.");*)
			       "1box"^(string_of_int (List.length info.succ))^"("^(up_point dt dd)^
				 (
				   match List.length info.succ with 
				     | 0 -> ")[l,t=#$"^s^"$#,b=0]"
				     | _ -> ","^(down_points info.time time_to_depth (time_to_depth.(info.time)) (info.succ))^")"^"[l,t=#$"^s^"$#,b=0]"
				 )
		   )
		 )
		 (*	     print_string "time_to_depth = [";
			     for i=0 to ((Array.length time_to_depth) - 2) do
			     print_int (time_to_depth.(i));
			     print_string ","
			     done;
			     print_endline ((string_of_int (time_to_depth.((Array.length (time_to_depth)) - 1)))^"]");*)
	   )
	)
	h;
      v
	
let vector_to_string v = 
  let l = List.map Array.to_list (Array.to_list v) in
    List.fold_right (fun a b -> a^"\\\\\n"^b) (List.map (fun x -> List.fold_right (fun a b -> a^"&"^b) x "") l) ""
