

(* General definitions.
 * 
 * We abstract the action of a strategy on a play by a closure operator.
 * No notion of Player and Opponent here, just parity of plays.
 *)

type polarity = Player | Opponent

type position = Left | Right

type ar_part = L | R | C

type label = Question of int | Answer of int

let opp = function 
  | Player -> Opponent 
  | Opponent -> Player

type move = {
  mutable previous : move option;   (* Previous move in a play *)
  mutable justifier : move option;  (* Justifying move in the play *)
  mutable h_justifier : move option;(* First visible hereditary justifier *)
  mutable work : move option;       (* Working pointer, used to calculate P-views *)
  mutable up_clone : move option;
  mutable downleft_clone : move option;
  mutable downright_clone : move option; (* In composite interaction, it is useful to keep track of cloned moves *)
  label : label;                      (* At each level, for each component, an unique identifiant *)
  component : int                   (* A play is separated into components *)
}
(*
let show_move s m = 
output_string s "	{\n";
(match m.previous with 
	| None -> output_string s "previous = None\n"
	| Some _ -> output_string s "previous = Some\n"
);
(match m.justifier with 
	| None -> output_string s "justifier = None\n"
	| Some _ -> output_string s "justifier = Some\n"
);
(match m.h_justifier with 
	| None -> output_string s "h_justifier = None\n"
	| Some _ -> output_string s "h_justifier = Some\n"
);
(match m.work with 
	| None -> output_string s "work = None\n"
	| Some _ -> print_endline "work = Some\n"
);
(match m.up_clone with 
	| None -> output_string s "up_clone = None\n"
	| Some _ -> output_string s "up_clone = Some\n"
);
(match m.downleft_clone with 
	| None -> output_string s "downleft_clone = None\n"
	| Some _ ->  output_string s "downleft_clone = Some\n"
);
(match m.downright_clone with 
	| None -> output_string s "downright_clone = None\n"
	| Some _ -> output_string s "downright_clone = Some\n"
);
(match m.label with 
	| Question i -> output_string s ("label = Question "^(string_of_int i)^"\n")
	| Answer k -> output_string s ("label = Answer "^(string_of_int k)^"\n")
);
output_string s ("component = "^(string_of_int m.component)^"\n");
output_string s " }\n"
*)

let show_move _ _ = ()



type play = {
  mutable current : move option; (* We only need to keep track of the last move *)
}

exception Empty_play

let h_channel p component = 
  let rec hchanworker m component last_found = 
    if m.component < component then
      last_found
    else
      match m.justifier with 
	| None -> 
	    if m.component = component
	    then last_found
	    else failwith "h_channel : no move found in this component"
	| Some j -> hchanworker j component m
  in match 
      (match p.current with 
	 | None -> failwith "h_channel : empty play !"
	 | Some m -> hchanworker m component m).label with 
	| Question q -> q
	| Answer k -> assert false
	    


(* Initialization and bookkeeping functions *)

let new_move l = {
  previous = None;
  justifier = None;
  h_justifier = None;
  work = None;
  up_clone = None;
  downleft_clone = None;
  downright_clone = None;
  label = Question l;
  component = 0
}

let new_play m = 
  {
    current = m
  }

let clear_work p = 
  let rec clear_worker m = 
    match m.work with 
      | None -> (match m.previous with 
		   | None -> ()
		   | Some (m') -> clear_worker m')
      | Some(_) -> 
	  m.work <- None; 
	  (match m.previous with 
	     | None -> ()
	     | Some (m') -> clear_worker m')
  in match p.current with 
    | None -> ()
    | Some x -> clear_worker x


let length p = 
  let rec aux m = match m.previous with
    | None -> 1
    | Some(m') -> 1 + aux m'
  in match p.current with 
    | None -> 0
    | Some x -> aux x



(* View extraction 
 * extract takes a play in argument. It walks through the play, and places the working pointers in order to
 * cover the hidden parts of the play. Then it returns the initial move of the thread. An innocent strategy
 * is now able to walk the view bottom-up via following the working pointers. 
 * 
 * Note : the initial move pointed by the outcome of extract is the first initial move of the thread
 * matching the prefix *)

exception Out_of_domain

(* val extract : int -> play -> move *)
let extract root play = 
  (* first step : walk down the pointers, setting the work pointers as inverses of the current path 
   * second step : climb up the work pointers, localize the first move matching the prefix, 
   *               remind its polarity as Opponent. Continue to climb up to find out the polarity of 
   *               the current move 
   * third stop : given the polarity of the current move, climb down the P-view, setting the work pointers
   *              to inverses of the path 
   * final step : return the initial move matching the prefix (it is now easy to clim up the P-view using
   *              the work pointers
   *)
(* val step1 : move -> move *)
  let rec step1 m = match m.justifier with (* returns the play initial move *)
    | None -> m
    | Some(m') -> 
	m'.work <- Some(m);
	step1 m'
(* val step2 : position -> ar_part -> move option -> move -> move -> polarity option -> move*polarity *)
  and step2 root initial_move final_move current_move pol = 
    (* returns the initial move of the restriction of the thread to the prefix, along with the polarity
     * of the last move *)
    match pol with 
      | None -> 
	  (
	    match current_move.component = root with 
	      | true -> 
		  step2 root (Some current_move) final_move current_move (Some Opponent)
	      | false -> 
		  (
		    match current_move.work with 
		      | None -> 
			  (
			    match current_move == final_move with
			      | true ->
				  raise Out_of_domain
			      | false -> 
				  failwith "extract:step2 : a work pointer is not set !"
			  )
		      | Some(m) -> 
			  step2 root None final_move m None
		  )
	  )
      | Some(p) -> 
	  (
	    match current_move.component - root <= 1 with 
	      | true -> 
		  (
		    match current_move.work with
		      | None -> 
			  ( 
			    match current_move == final_move with 
			      | true -> 
				  (
				    match initial_move with 
				      | None -> 
					  failwith "extract:step2 : final move reached and no initial move is set ? "
				      | Some(i) -> (i,p)
				  )
			      | false -> 
				  failwith "extract:step2 : A work pointer is not set !"
			  )
		      | Some(m) -> 
			  step2 root initial_move final_move m (Some (opp p))
		  )
	      | false ->
		  raise Out_of_domain
	  )
	  (* val step3 : move -> move -> polarity -> unit *)
  and step3 initial_move current_move pol = 
    match (current_move==initial_move) with
      | true -> ()
      | false -> ( match pol with 
		     | Player -> ( match current_move.previous with 
				     | None -> failwith "extract:step3: Beginning of the play reached without seeing the expected initial move ! (btw, first move is a Player move)"
				     | Some(m) -> 
					 m.work <- Some(current_move);
					 step3 initial_move m Opponent )
		     | Opponent -> ( match current_move.justifier with
				       | None -> failwith "extract:step3: Beginning of the play reached without seeing the expected initial move !"
				       | Some(m) -> 
					   m.work <- Some(current_move);
					   step3 initial_move m Player ))
  in match play.current with 
    | None -> raise Empty_play
    | Some m -> let thread_initial = step1 m
      in let (prefix_initial,pol) = step2 root None m thread_initial None
      in let _ = step3 prefix_initial m pol 
      in prefix_initial 
  


let add_move j lab p poly curry =
  let initial m c = 
    match m.justifier with 
      | None -> true
      | Some k when k.component < c -> true
      | _ -> false 
  and visible x = x.component = 0 || x.component = 2 in
  let comp = match curry with 
    | 0 -> 1
    | _ -> 0 in  
  let (label,position) = 
    if (initial j comp) then 
      (
	match lab with 
	  | Question q -> 
	      (
		match curry with 
		  | 0 -> 
		      if q < poly then 
			(lab,Left)
		      else
			(Question (q-poly), Right)
		  | _ -> 
		      if
			q < poly + curry 
		      then 
			(Question q,Left)
		      else
			(Question (q-curry),Right)
	      )
	  | Answer k -> (Answer k, Right)
      )
    else
      (
	if j.component > comp then 
	  (lab, Left)
	else
	  (lab, Right)
      )
  in 
  let m = 
    {
      previous = p.current;
      justifier = Some j;
      h_justifier = if (visible j) then Some j else j.h_justifier;
      work = None;
      up_clone = None;
      downleft_clone = None;
      downright_clone = None;
      component = (match position with Left -> comp + 1 | Right -> comp);
      label = label
    } in 
    p.current <- Some m;
    position
      
      
    
    

