(* Definition of interactions, basic operations *)

open Games


(* We first define the notion of agent, which can be either a simple closure, or a whole interaction *)

(* closure = play -> root -> int *)

type ask_action =
  | ABusy         (* Computation in progress *)
  | ALeft         (* Output on the Left component *)
  | ARight        (* Output on the Right component *)
  | ANone         (* Agent added no information *)
  | ADeadlock     (* Agent is deadlocked *)

let string_of_action = function
  | ABusy -> "Busy"
  | ALeft -> "Left"
  | ARight -> "Right"
  | ANone -> "Nothing"
  | ADeadlock -> "Deadlocked"



type closure = {
  c_arity : int;
  close : int -> int -> play -> int -> int -> int -> ask_action
}



(* close poly curry p component dec last_component :
 *    Applies the closure:
 *            - with poly free variables
 *	      - with curry arguments
 *            - on the play p
 *            - rooted at component
 *            - its input is shifted by dec
 *            - if we are to play in last_component
 *              we know that we must handle h_justifier differently
 *)


type status = 
  | InTau of int  (* k-th agent is to play next *)
  | Ready_i       (* Agent is ready, waiting for input *)
  | Ready_o       (* Agent is ready, waiting for output *)
  | Deadlocked    (* Agent is deadlocked *)


let string_of_status = function
  | InTau _ -> "Busy"
  | Ready_i -> "Waiting Input"
  | Ready_o -> "Waiting Output"
  | Deadlocked -> "Deadlock"


type agent = Closure of closure | Interaction of interaction | Tensor of agent*agent

and strategy = 
    {
      agent : agent;
      curry : int
    }

and interaction = 
    {
      arity : int;
      play : Games.play;
      strategies : strategy array;
      mutable status : status;
      poly : int
    }

let string_of_pos = function 
	| Left -> "L"
	| Right -> "R"


let rec string_of_agent = function
  | Closure _ -> "C"
  | Interaction i -> 
  	let k = match i.status with 
		| Ready_i -> "i"
		| Ready_o -> "o"
		| InTau k -> string_of_int k 
		| _ -> ";" in
      let n = Array.length i.strategies in
      let s = ref ("["^(string_of_agent (i.strategies.(n-1).agent))) in
	for j = 1 to n-1 do
	  s := ((!s)^k^(string_of_agent (i.strategies.(n-1-j).agent)))
	done;
	s := !s^"]";
	!s
  | Tensor(a,b) -> 
      ("("^(string_of_agent a)^"⊗"^(string_of_agent b)^(")"))
			
	  

let rec status = function 
  | Closure _ -> Ready_o
  | Interaction i -> i.status
  | Tensor (ag1,ag2) -> 
      (
	match (status ag1, status ag2) with 
	  | Ready_i , Ready_i -> Ready_i
	  | Ready_o , Ready_o -> Ready_o
	  | Ready_i , Ready_o
	  | Ready_o , Ready_i -> Ready_i
	  | InTau _ , _ 
	  |  _      , InTau _ -> InTau (-1) (* InTau -1 just mean Busy, in an unspecified location *)
	  | Deadlocked , _
	  | _ , Deadlocked -> Deadlocked 
		
      )

(* We define two recursive functions to extract the arity of an agent *)

let rec arity = function 
  | Closure c -> c.c_arity
  | Interaction i -> i.arity
  | Tensor(ag1,ag2) -> (arity ag1) + (arity ag2)


(* We define two recursive functions :
 *
 *  - step : scans the agent until it finds some step of computation
 *  - ask p : + first check that the agent has no computation running.
 *            + check the last move of p
 *                       - it is cloned, check for an answer
 *                                      + there is an answer -> copy it
 *                                      + otherwise -> deadlock
 *                       - otherwise, clone it, set the status to Busy
 *)


let rec step ag = 
  Debug.debug ("Stepping agent "^(string_of_agent ag));
  match ag with 
    | Closure _ -> Ready_o
    | Tensor(ag1,ag2) ->
	( 
	  match status ag1 with
	    | InTau _ -> step ag1
	    | Deadlocked -> Deadlocked
	    | Ready_i | Ready_o -> 
		(
		  match status ag2 with 
		    | InTau _ -> step ag2
		    | x -> x
		)
	)
    | Interaction i -> 
	(
	  match i.status with 
	    | InTau k -> 
		(
		  match 
		    (*Debug.debug ("step: asking "^(string_of_agent (Interaction i)));*)
		    step (i.strategies.(k).agent) 
		  with 
		    | InTau _ as a-> a
		    | Deadlocked -> Deadlocked
		    | Ready_i | Ready_o -> 
			(
			  match 
			
			    ask i.poly i.strategies.(k).curry i.play i.strategies.(k).agent k 0 (Array.length i.strategies)

			  with 
			    | ABusy -> 
				InTau(-1)
			    | ALeft -> 
				Debug.debug "Left !";
				(
				  match k+1=Array.length i.strategies with 
				    | true -> 
					i.status <- Ready_o;
					Ready_o
				    | false -> 
					i.status <- InTau (k+1);
					InTau (-1)
				)
			    | ARight -> 
				Debug.debug "Right !";
				(
				  match k with 
				    | 0 -> 
					i.status <- Ready_o;
					Ready_o
				    | _ -> 
					i.status <- InTau (k-1);
					InTau (-1)
				)
			    | ADeadlock -> Deadlocked
			    | ANone -> Deadlocked
			)
		)
	    | Ready_i -> Ready_i
	    | Ready_o -> Ready_o (*theoretically, we never step a finished computation... *)
            | Deadlocked -> Deadlocked 
	)



and ask poly curry p agent component dec last_component = 
  Debug.debug ("Asking agent "^(string_of_agent agent));
  match p.current with 
    | None -> ANone
    | Some current_p ->
	(
	  match agent,(status agent) with 
	      
	    | Closure c , _ -> 
	    	(
	    	        c.close poly curry p component dec last_component
		)
	    | Tensor (ag1,ag2) , _ ->
		
		  Debug.debug "ask: asking a tensor.";
		  let channel = h_channel p component in
		    Debug.debug ("ask: play is on channel "^(string_of_int channel));
		    Debug.debug ("ask: aftter dec, the channel is "^(string_of_int (channel-dec)));
		    let ar1 = arity ag1
		    and ar2 = arity ag2 in
		      if 
			(channel-dec) < ar1
		      then 
			(			
			  Debug.debug "ask: the play seem to be addressed to ag1";
			  let action = ask poly curry p ag1 component dec last_component 
			  in Debug.debug (string_of_action action);
			    action
			)
		      else
			if 
			  (
			    Debug.debug ("ar1 = "^(string_of_int ar1)^" and ar2 = "^(string_of_int ar2));
			    (channel - dec - ar1) < ar2
			  )
			then 
			  (
			    Debug.debug "ask: the play seem to be adressed to ag2";
			    ask poly curry p ag2 component (dec + ar1) last_component
			  )
			else 
			  failwith "ask: the play seems out of domain !"
		  
	    | Interaction i , Ready_i -> 
		(* We're about to copy an outer new move into the inner interaction.
		 * TODO: we should check that the last move is really to be copied :
		 * check that one of p.current.downleft_clone and p.current.downright_clone
		 * is set to None *)
		(
		  match p.current with 
		    | None -> assert false
		    | Some current_p -> 
			
			let previous = 
			  match i.play.current with 
			    | None -> None
			    | Some m -> Some m
			and pos_agent = 
			  match component with 
			    | 0 -> Games.Right
			    | 1 -> Games.Left
			    | _ -> failwith "n-ary interaction not supported"
			in let justifier =
			  match current_p.justifier with  
			    | None -> None
			    | Some m -> 
				( 
				  match pos_agent with 
				    | Left -> 
					(
					  match m.downleft_clone with 
					    | None -> None (* new move is initial in the restriction *)
					    | Some m' -> Some m'
					)
				    | Right ->
					(
					  match m.downright_clone with 
					    | None -> failwith "ask:Read_i: I am the right member of the interaction and I see a justifier with no downright_clone !"
					    | Some m' -> Some m'
					)
				) in
			let position =  
			  (
			    match justifier with 
			      | None -> Games.Right
			      | Some j -> 
				  (
				    match j.component with 
				      | 0 -> Games.Right
				      | _ -> Games.Left
				  )
			  )
			in let h_justifier = justifier (* an outer move cannot be justified by a hidden move *)
			   and label = (
			     if justifier = None then 
			       match current_p.label with 
				 | Question i -> Question (i-dec)
				 | Answer _ -> failwith "ask:Read_i : just encountered an initial answer"
			     else current_p.label
			   )
			   and comp = (
			     match position with
			       | Games.Left -> 2
			       | Games.Right -> 0
			   )
			in let new_move = {
			    previous = previous;
			    justifier = justifier;
			    h_justifier = h_justifier;
			    work = None;
			    up_clone = p.current;
			    downleft_clone = None;
			    downright_clone = None;
			    label = label;
			    component = comp
			  }
			in 
			  Debug.debug ("I am "^(string_of_agent (Interaction i)));
			  Debug.debug "I am inputing : ";
			  Games.show_move stderr new_move;
			  (
			    match pos_agent with 
			      | Left -> 
				  current_p.downleft_clone <- Some new_move
			      | Right -> 
				  current_p.downright_clone <- Some new_move
			  );
			  (
			    match position with 
			      | Left -> 
				  i.status <- InTau(1)
			      | Right -> 
				  i.status <- InTau(0)
			  );
			  i.play.current <- Some new_move;
			  ABusy
		)
	    | Interaction i , Ready_o -> 
		(* We're about to copy an inner new move to the outer interaction.
		 * TODO: check for safety that i.play.current.up_clone is set to None *)
		(
		  match p.current with 
		    | None -> assert false
		    | Some current_p -> 
			(
			  match i.play.current with 
			    | None -> ANone
			    | Some i_play_current -> 
				let pos_strat = 
				  match component with 
				    | 0 -> Right
				    | _ -> Left in
				let previous = p.current 
				and justifier = 
				  (
				    match i_play_current.h_justifier with
				      | None -> failwith "ask: an internal move needs to be justified"
				      | Some(m) -> 
					  ( 
					    match m.up_clone with 
					      | None -> failwith "ask: the hereditary justifier should be up_cloned !! "
					      | Some(c) -> Some(c)
					  )
				  )
				and h_justifier = 
				  (
				    match i_play_current.h_justifier with 
				      | None -> failwith "ask: A P-move has no h_justifier !"
				      | Some m -> 
					  (
					    match m.up_clone with 
					      | None -> failwith "ask: the h_justifier has no up_clone !"
					      | Some x as c -> 
						  (
						    match x.component with 
						      | 0 -> c
						      | n when n = last_component -> c 
						      | _ -> x.h_justifier
						  )
					  )
				  )
				    
				in let (label,pos) =
				    match i_play_current.h_justifier with 
				      | None -> assert false (* Some closure has played an unjustified move *)
				      | Some j -> 
					  (
					    match j.justifier with 
					      | None -> 
						  (
						    match i_play_current.label with
						      | Question q ->
							  (
							    match curry with
							      | 0 ->
								  if
								    (
								      Debug.debug ("Left position. poly = "^(string_of_int poly)^" and i= "^(string_of_int q));
								      q<poly
								    )
								  then
								    (Question q,Left)
								  else
								    (Question (q-poly),Right)
							      | n ->
								  if
								    (
								      Debug.debug ("q = "^(string_of_int q)^" , poly = "^(string_of_int poly));
								      q < poly + n
								    )
								  then
								    (Question q,Left)
								  else
								    (Question (q-curry),Right) (* q - poly - 1 *)
							  )
						      | Answer k -> (Answer k,Games.Right)
						  )
					      | Some q -> 
						  (
						    match justifier with 
						      | None -> assert false
						      | Some justifier -> 
							  if justifier.component = component
							  then 
							    (i_play_current.label, Right)
							  else 
							    (i_play_current.label, Left)
						  )
					  )
				in let comp = match pos with 
				  | Left -> component + 1
				  | Right -> component
				      
				and downleft_clone = 
				  ( 
				    match pos_strat with 
				      | Games.Left -> i.play.current
				      | Games.Right -> None
				  )
				and downright_clone = 
				  (
				    match pos_strat with 
				      | Games.Left -> None
				      | Games.Right -> i.play.current
				  )
				in let new_move = {
				    previous = previous;
				    justifier = justifier;
				    h_justifier = h_justifier;
				    work = None;
				    up_clone = None;
				    downleft_clone = downleft_clone;
				    downright_clone = downright_clone;
				    label = label;
				    component = comp
				  }
				in 
			  	  Debug.debug ("I am "^(string_of_agent (Interaction i)));
			  	  Debug.debug "I am outputing : ";
			  	  Games.show_move stderr new_move;
				  i_play_current.up_clone <- Some new_move;
				  p.current <- Some new_move;
				  i.status <- Ready_i;
				  (
				    match pos with 
				      | Games.Left -> ALeft
				      | Games.Right -> ARight
				  )
			)
		)
	    | _ , InTau _ -> ABusy
	    | _ , Deadlocked -> ADeadlock
	)




let new_interaction l m = 
  let v = Array.of_list (List.rev l) in 
  let ar = arity v.(0).agent in
  let p = Games.new_play m 
  and stat = 
    match m with 
      | None -> Ready_i
      | Some _ -> InTau 0 in 
    { 
      arity = ar;
      play = p;
      strategies = v;
      status = stat;
      poly = 0
    }

    
